
#include <string.h>

#include "module_utils.h"

#include "label_generator.h"
#include "text_reader.h"
#include "pac_logger.h"

#define LANG_EN 1
#define LANG_FA 2

PQueue *label_queue ;
size_t pactos_audio_chunk_size ;

Flite_Text_Analyzer analyzer_en;

pthread_t label_gen_thread;
void *_generate_label(void *nothing);


gboolean free_label(char **p_labels, size_t size)
{
	for(size_t i = 0; i < size; i++)
	{
		g_free(p_labels[i]);
	}
	g_free(p_labels);
	p_labels = NULL;

	return TRUE;
}

static char **allocate_label(size_t size)
{
	size_t label_len = 250;
	char **p_labels = g_new(char *, size);
	if(p_labels == NULL)
	{
		return NULL;
	}
	for(size_t i = 0; i < size; i++)
	{
		p_labels[i] = g_new(char, label_len);
		if(p_labels[i] == NULL)
		{
			if(i>0)
			{
				free_label(p_labels, i);
			}
			return NULL;
		}
		memset(p_labels[i],0,label_len);
	}
	
	return p_labels;
}

char **label_dup(char **src, size_t size)
{
	char **new_label = allocate_label(size);
	for(size_t i = 0; i < size; i++)
		strcpy(new_label[i], src[i]);

	return new_label;
}

static void log_label(char **labels, size_t n)
{	
	PAC_LOG("-----------------------------------------------------------------------");
	for(size_t i = 0; i < n; i++)
	{
		PAC_LOG("%s", labels[i]);
	}
	PAC_LOG("-----------------------------------------------------------------------");
}

gboolean label_generator_init()
{
	
	if(espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS, pactos_audio_chunk_size,
				 "/usr/lib/arm-linux-gnueabihf", 0) < 0)
	{
		PAC_LOG("Can not initialize espeak!!!\n");
		return FALSE;
	}
	if(espeak_SetVoiceByName("fa") < 0)
	{
		PAC_LOG("can not set voice by name\n");
		return FALSE;
	}
	
	int ret = pthread_create(&label_gen_thread, NULL, _generate_label, NULL);
	if(ret != 0)
	{
		PAC_LOG("Can't create label generator thread!!!");
		return FALSE;
	}
	
	label_queue = pactos_queue_create();

	return TRUE;
}

void label_generator_close()
{
	espeak_Terminate();
	
	pthread_cancel(label_gen_thread);
	pthread_join(label_gen_thread, NULL);

	pactos_queue_clear(label_queue);
}

size_t get_hts_label(void *text, void **out_text, gboolean first_clause, char ***hts_labels)
{
	int num_labels, label_len;
	int msg_len = strlen((char *)text);
	label_len = 250;	
	num_labels = msg_len + msg_len / 2 + 20;
	
	*hts_labels = allocate_label(num_labels);
	if(*hts_labels == NULL)
	{
		PAC_LOG("Can't allocate memory for hts labels!!!");
		return 0;
	}
	
	size_t num_generated = get_part_hts_labels(text, 
						   out_text,
						   first_clause, 
						   *hts_labels,
						   num_labels,
						   label_len,
						   0);
	if(num_generated <= 0)
	{
		free_label(*hts_labels, num_labels);
	}

	return num_generated;
}


void *_generate_label(void *nothing)
{
	void *crr_text = NULL;
	void *p_text = NULL;
	
	GArray *splitted_text = NULL;

	set_speaking_thread_parameters();

	while(1)
	{
		splitted_text = pactos_queue_pop(text_queue);
		for(size_t part_idx = 0; part_idx < splitted_text->len; part_idx++)
		{
			GString *crr_part = g_array_index(splitted_text, GString *, part_idx);
			crr_text = (void *)crr_part->str;
			p_text = crr_text;
			void *out_text = p_text;
			gboolean first_clause = TRUE;
			char **hts_labels, **en_labels;


			do
			{

				size_t num_generated = 0;
				char lang = get_text_script(p_text); 
				if(lang == LANG_FA)
				{

					num_generated = get_hts_label(p_text, 
								     &out_text,
								     first_clause, 
								     &hts_labels);
				}
				else
				{
					Flite_Text_Analyzer_initialize(&analyzer_en);
					Flite_Text_Analyzer_analysis(&analyzer_en, p_text);
					Flite_Text_Analyzer_get_label_data(&analyzer_en, &en_labels, 
									   &num_generated);
					hts_labels = label_dup(en_labels, num_generated);
					free_label(en_labels, num_generated);
					Flite_Text_Analyzer_clear(&analyzer_en);
					out_text = NULL;
				}

				if(num_generated <= 0)
				{
					PAC_LOG("Can't generate hts labels!!!");
					break;				
				}
			
				// Do whatever you want with hts labels
	//			log_label(hts_labels, num_generated);
				HLabel *crr_label = g_new(HLabel, 1);
				crr_label->label = hts_labels;
				crr_label->size = num_generated;
				crr_label->lang = lang;
				crr_label->event_begin = (part_idx==0)?TRUE:FALSE;
				crr_label->event_end = (part_idx==splitted_text->len-1)?TRUE:FALSE;
				
//				PAC_LOG("lang == %d", (int)lang);
//				PAC_LOG("size == %lu", crr_label->size);
//				log_label(crr_label->label, crr_label->size);
				pactos_queue_push(label_queue, crr_label);

				if(first_clause)
				{
					first_clause = FALSE;
				}
			
				p_text = out_text;
			
			}while(p_text);	
		
			g_string_free(crr_part, TRUE);
			crr_text = NULL;	
		}
		g_array_free(splitted_text, FALSE);
	}	

}




