
/*
 * espeak.c - Speech Dispatcher backend for espeak
 *
 * Copyright (C) 2007 Brailcom, o.p.s.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * @author Lukas Loehrer
 * Based on ibmtts.c.
 *
 * $Id: espeak.c,v 1.11 2008-10-15 17:04:36 hanke Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* < Includes*/

/* System includes. */
#include <string.h>
#include <glib.h>
#include <semaphore.h>

/* required header file */

#include <text_reader.h>
#include <label_generator.h>
#include <synthesizer.h>
#include <player.h>
#include <pac_logger.h>

#ifndef ESPEAK_API_REVISION
#define ESPEAK_API_REVISION 1
#endif

/* Speech Dispatcher includes. */
#include "spd_audio.h"
#include <speechd_types.h>
#include "module_utils.h"

/* > */
/* < Basic definitions*/

#define MODULE_NAME     "pactos"
#define DBG_MODNAME     "Pactos:"

#define MODULE_VERSION  "1.0"

#define DEBUG_MODULE 1
DECLARE_DEBUG()

typedef enum {
	FATAL_ERROR = -1,
	OK = 0,
	ERROR = 1
} TPactosSuccess;

char **pactos_text_message = NULL;

extern pthread_mutex_t player_stop_mutex;
extern pthread_cond_t player_stop_condition;

 SPDVoice **pactos_voice_list = NULL;

/* <Function prototypes*/

/* Internal function prototypes for main thread. */

/* Basic parameters */
static void pactos_set_rate(signed int rate);
static void pactos_set_pitch(signed int pitch);
static void pactos_set_pitch_range(signed int pitch_range);
static void pactos_set_volume(signed int volume);
static void pactos_set_punctuation_mode(SPDPunctuation punct_mode);
static void pactos_set_cap_let_recogn(SPDCapitalLetters cap_mode);

/* Voices and languages */
static void pactos_set_language(char *lang);
static void pactos_set_voice(SPDVoiceType voice);
static void pactos_set_language_and_voice(char *lang, SPDVoiceType voice);
static void pactos_set_synthesis_voice(char *);


/* Miscellaneous internal function prototypes. */
static gboolean is_thread_busy(pthread_mutex_t * suspended_mutex);

/* The playback thread start routine. */


// Utility functions



/* > */
/* < Module configuration options*/

MOD_OPTION_1_STR(PactosPunctuationList)
MOD_OPTION_1_INT(PactosCapitalPitchRise)
MOD_OPTION_1_INT(PactosMinRate)
MOD_OPTION_1_INT(PactosNormalRate)
MOD_OPTION_1_INT(PactosMaxRate)
MOD_OPTION_1_INT(PactosListVoiceVariants)

MOD_OPTION_1_INT(PactosAudioChunkSize)
MOD_OPTION_1_INT(PactosAudioQueueMaxSize)
MOD_OPTION_1_STR(PactosSoundIconFolder)
MOD_OPTION_1_INT(PactosSoundIconVolume)


/* > */
/* < Public functions */
int module_load(void)
{
	INIT_SETTINGS_TABLES();

	REGISTER_DEBUG();

	/* Options */
	MOD_OPTION_1_INT_REG(PactosAudioChunkSize, 2000);
	MOD_OPTION_1_INT_REG(PactosAudioQueueMaxSize, 20 * 22050);
	MOD_OPTION_1_STR_REG(PactosSoundIconFolder,
			     "/usr/share/sounds/sound-icons/");
	MOD_OPTION_1_INT_REG(PactosSoundIconVolume, 0);

	MOD_OPTION_1_INT_REG(PactosMinRate, 80);
	MOD_OPTION_1_INT_REG(PactosNormalRate, 170);
	MOD_OPTION_1_INT_REG(PactosMaxRate, 390);
	MOD_OPTION_1_STR_REG(PactosPunctuationList, "@/+-_");
	MOD_OPTION_1_INT_REG(PactosCapitalPitchRise, 800);
	MOD_OPTION_1_INT_REG(PactosListVoiceVariants, 0);
	if (PactosCapitalPitchRise == 1 || PactosCapitalPitchRise == 2) {
		PactosCapitalPitchRise = 0;
	}

	return 0;
}

int module_init(char **status_info)
{
	int ret;
//	const char *pactos_version;

	DBG(DBG_MODNAME " Module init().");
	INIT_INDEX_MARKING();

	*status_info = NULL;
	
	
	if(!pac_logger_init())
	{
		fprintf(stderr, "Can' init module!!! (initializing logger failed!)");
		return -1;
	}
	
/*	if(!text_reader_init())
	{
		PAC_LOG("Can't init module!!! (initializing text reader failed!)");
		return -1;
	}*/
	text_queue = pactos_queue_create();
		
	if(!label_generator_init())
	{
		PAC_LOG("Can't init module!!! (initializing label generator failed!)");
		return -1;
	}
	
	if(!synthesizer_init())
	{
		PAC_LOG("Can't init module!!! (initializing synthesizer failed!)");
		return -1;
	}
	
	if(!player_init())
	{
		PAC_LOG("Can't init module!!! (initializing player failed!)");
		return -1;
	}

	pactos_text_message = g_malloc(sizeof(char *));
	*pactos_text_message = NULL;

	*status_info = g_strdup(DBG_MODNAME " Initialized successfully.");

	return 0;
}

SPDVoice **module_list_voices(void)
{
	return pactos_voice_list;
}

int module_speak(gchar * data, size_t bytes, SPDMessageType msgtype)
{

	DBG(DBG_MODNAME " Requested data: |%s| %d %lu", data, msgtype, (unsigned long)bytes);
	PAC_LOG(" Requested data: |%s| %d %lu", data, msgtype, (unsigned long)bytes);
	
	/* Setting speech parameters. */
	UPDATE_STRING_PARAMETER(voice.language, pactos_set_language);
	UPDATE_PARAMETER(voice_type, pactos_set_voice);
	UPDATE_STRING_PARAMETER(voice.name, pactos_set_synthesis_voice);

	UPDATE_PARAMETER(rate, pactos_set_rate);
	UPDATE_PARAMETER(volume, pactos_set_volume);
	UPDATE_PARAMETER(pitch, pactos_set_pitch);
	UPDATE_PARAMETER(pitch_range, pactos_set_pitch_range);
	UPDATE_PARAMETER(punctuation_mode, pactos_set_punctuation_mode);
	UPDATE_PARAMETER(cap_let_recogn, pactos_set_cap_let_recogn);
	PAC_LOG("after usettings");


/*	if(*pactos_text_message != NULL)
	{
		PAC_LOG("module_speak : in if before g_free message");
		g_free(*pactos_text_message);
		PAC_LOG("module_speak : in if after g_free message");
		*pactos_text_message = NULL;
	}
*/

	*pactos_text_message = module_strip_ssml(data);
	PAC_LOG("module_speak : message is |%s|", *pactos_text_message);

	GArray *splitted_text = split_text_to_langs(*pactos_text_message);
	pactos_queue_push(text_queue, splitted_text);
	
	g_free(*pactos_text_message);	
	//sem_post(&text_reader_semaphore);

	DBG(DBG_MODNAME " Leaving module_speak() normally.");
	return TRUE;
}

int module_stop(void)
{
	DBG(DBG_MODNAME " module_stop().");
	PAC_LOG("module_stop() ...");
	sem_post(&player_stop_semaphore);

//	pactos_queue_clear(text_queue);
//	pactos_queue_clear(label_queue);
//	pactos_queue_clear(sample_queue);
	
	return 0;
}

size_t module_pause(void)
{
	DBG(DBG_MODNAME " module_pause().");
	PAC_LOG("module_pause()");

	
	
	return 0;
}

int module_close(void)
{
	DBG(DBG_MODNAME " close().");
	PAC_LOG("module_close()");

	DBG(DBG_MODNAME " Terminating threads");
	
	pac_logger_close();
	
	text_reader_close();
	
	label_generator_close();
	
	synthesizer_close();
	
	player_close();

	return 0;
}

/* > */
/* < Internal functions */
/* Return true if the thread is busy, i.e., suspended mutex is not locked. */
static gboolean is_thread_busy(pthread_mutex_t * suspended_mutex)
{
	if (EBUSY == pthread_mutex_trylock(suspended_mutex))
	{
		return FALSE;
	}
	else
	{
		pthread_mutex_unlock(suspended_mutex);
		return TRUE;
	}
}



static void pactos_set_rate(signed int rate)
{
	
	PAC_LOG("SET_RATE : rate = %d", rate);
	assert(rate >= -100.0 && rate <= 100.0);

	signed int old_range = 200;
	signed int new_range = 2;

	float new_value = (((rate + 100.0) * new_range) / old_range) + 0;
	
	HTS_Engine_set_speed(&engine_fa, new_value);
	HTS_Engine_set_speed(&engine_en, new_value);

//	int speed;
//	int normal_rate = PactosNormalRate, max_rate = PactosMaxRate, min_rate =
//	    PactosMinRate;
//
//	if (rate < 0)
//		speed = normal_rate + (normal_rate - min_rate) * rate / 100;
//	else
//		speed = normal_rate + (max_rate - normal_rate) * rate / 100;
//
//	espeak_ERROR ret = espeak_SetParameter(espeakRATE, speed, 0);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Error setting rate %i.", speed);
//	} else {
//		DBG(DBG_MODNAME " Rate set to %i.", speed);
//	}
}

static void pactos_set_volume(signed int volume)
{
//	assert(volume >= -100 && volume <= +100);
//	int vol;
//	vol = volume + 100;
//	espeak_ERROR ret = espeak_SetParameter(espeakVOLUME, vol, 0);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Error setting volume %i.", vol);
//	} else {
//		DBG(DBG_MODNAME " Volume set to %i.", vol);
//	}
}

static void pactos_set_pitch(signed int pitch)
{
//	assert(pitch >= -100 && pitch <= +100);
//	int pitchBaseline;
//	/* Possible range 0 to 100. */
//	if (pitch < 0) {
//		pitchBaseline =
//		    ((float)(pitch + 100) * pactos_voice_pitch_baseline) /
//		    (float)100;
//	} else {
//		pitchBaseline =
//		    (((float)pitch * (100 - pactos_voice_pitch_baseline))
//		     / (float)100) + pactos_voice_pitch_baseline;
//	}
//	assert(pitchBaseline >= 0 && pitchBaseline <= 100);
//	espeak_ERROR ret = espeak_SetParameter(espeakPITCH, pitchBaseline, 0);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Error setting pitch %i.", pitchBaseline);
//	} else {
//		DBG(DBG_MODNAME " Pitch set to %i.", pitchBaseline);
//	}
}

static void pactos_set_pitch_range(signed int pitch_range)
{
//	assert(pitch_range >= -100 && pitch_range <= +100);
//	int pitchRangeBaseline;
//	/* Possible range 0 to 100. */
//	if (pitch_range < 0) {
//		pitchRangeBaseline =
//		    ((float)(pitch_range + 100) *
//		     pactos_voice_pitch_range_baseline) / (float)100;
//	} else {
//		pitchRangeBaseline =
//		    (((float)pitch_range *
//		      (100 - pactos_voice_pitch_range_baseline))
//		     / (float)100) + pactos_voice_pitch_range_baseline;
//	}
//	assert(pitchRangeBaseline >= 0 && pitchRangeBaseline <= 100);
//	espeak_ERROR ret =
//	    espeak_SetParameter(espeakRANGE, pitchRangeBaseline, 0);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Error setting pitch range %i.",
//		    pitchRangeBaseline);
//	} else {
//		DBG(DBG_MODNAME " Pitch range set to %i.", pitchRangeBaseline);
//	}
}

static void pactos_set_punctuation_mode(SPDPunctuation punct_mode)
{
//	espeak_PUNCT_TYPE espeak_punct_mode = espeakPUNCT_SOME;
//	switch (punct_mode) {
//	case SPD_PUNCT_ALL:
//		espeak_punct_mode = espeakPUNCT_ALL;
//		break;
//	case SPD_PUNCT_SOME:
//		espeak_punct_mode = espeakPUNCT_SOME;
//		break;
//	case SPD_PUNCT_NONE:
//		espeak_punct_mode = espeakPUNCT_NONE;
//		break;
//	}
//
//	espeak_ERROR ret =
//	    espeak_SetParameter(espeakPUNCTUATION, espeak_punct_mode, 0);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Failed to set punctuation mode.");
//	} else {
//		DBG("Set punctuation mode.");
//	}
}

static void pactos_set_cap_let_recogn(SPDCapitalLetters cap_mode)
{
//	int espeak_cap_mode = 0;
//	switch (cap_mode) {
//	case SPD_CAP_NONE:
//		espeak_cap_mode = PactosCapitalPitchRise;
//		break;
//	case SPD_CAP_SPELL:
//		espeak_cap_mode = 2;
//		break;
//	case SPD_CAP_ICON:
//		espeak_cap_mode = 1;
//		break;
//	}
//
//	espeak_ERROR ret =
//	    espeak_SetParameter(espeakCAPITALS, espeak_cap_mode, 1);
//	if (ret != EE_OK) {
//		DBG(DBG_MODNAME " Failed to set capitals mode.");
//	} else {
//		DBG("Set capitals mode.");
//	}
}

/* Given a language code and SD voice code, sets the espeak voice. */
static void pactos_set_language_and_voice(char *lang, SPDVoiceType voice_code)
{
	DBG(DBG_MODNAME " set_language_and_voice %s %d", lang, voice_code);
	espeak_ERROR ret;

	unsigned char overlay = 0;
	switch (voice_code) {
	case SPD_MALE1:
		overlay = 0;
		break;
	case SPD_MALE2:
		overlay = 1;
		break;
	case SPD_MALE3:
		overlay = 2;
		break;
	case SPD_FEMALE1:
		overlay = 11;
		break;
	case SPD_FEMALE2:
		overlay = 12;
		break;
	case SPD_FEMALE3:
		overlay = 13;
		break;
	case SPD_CHILD_MALE:
		overlay = 4;
		break;
	case SPD_CHILD_FEMALE:
		overlay = 14;
		break;
	default:
		overlay = 0;
		break;
	}

	char *name = g_strdup_printf("%s+%d", lang, overlay);
	DBG(DBG_MODNAME " set_language_and_voice name=%s", name);
	ret = espeak_SetVoiceByName(name);

	if (ret != EE_OK) {
		DBG(DBG_MODNAME " Error selecting language %s", name);
	} else {
		DBG(DBG_MODNAME " Successfully set voice to \"%s\"", name);
	}
	g_free(name);
}

static void pactos_set_voice(SPDVoiceType voice)
{
//	assert(msg_settings.voice.language);
	//pactos_set_language_and_voice(msg_settings.voice.language, voice);
}

static void pactos_set_language(char *lang)
{
	//pactos_set_language_and_voice(lang, msg_settings.voice_type);
}

static void pactos_set_synthesis_voice(char *synthesis_voice)
{
//	if (synthesis_voice != NULL) {
//		espeak_ERROR ret = espeak_SetVoiceByName(synthesis_voice);
//		if (ret != EE_OK) {
//			DBG(DBG_MODNAME " Failed to set synthesis voice to %s.",
//			    synthesis_voice);
//		}
//	}
}


static SPDVoice **pactos_list_synthesis_voices()
{
	SPDVoice **result = NULL;
	SPDVoice *voice = NULL;
	SPDVoice *vo = NULL;
	const espeak_VOICE **espeak_voices = NULL;
	const espeak_VOICE **espeak_variants = NULL;
	espeak_VOICE *variant_spec = NULL;
	const espeak_VOICE *v = NULL;
	GQueue *voice_list = NULL;
	GQueue *variant_list = NULL;
	GList *voice_list_iter = NULL;
	GList *variant_list_iter = NULL;
	const gchar *first_lang = NULL;
	gchar *lang = NULL;
	gchar *variant = NULL;
	gchar *dash = NULL;
	gchar *vname = NULL;
	int numvoices = 0;
	int numvariants = 0;
	int totalvoices = 0;
	int i = 0;

	espeak_voices = espeak_ListVoices(NULL);
	voice_list = g_queue_new();

	for (i = 0; espeak_voices[i] != NULL; i++) {
		v = espeak_voices[i];
		if (!g_str_has_prefix(v->identifier, "mb/")) {
			/* Not an mbrola voice */
			voice = g_new0(SPDVoice, 1);

			voice->name = g_strdup(v->name);

			first_lang = v->languages + 1;
			lang = NULL;
			variant = NULL;
			if (g_utf8_validate(first_lang, -1, NULL)) {
				dash =
				       g_utf8_strchr(first_lang, -1, '-');
				if (dash != NULL) {
					/* There is probably a variant string (like en-uk) */
					lang =
					    g_strndup(first_lang,
						      dash - first_lang);
					variant =
					    g_strdup(g_utf8_next_char(dash));
				} else {
					lang = g_strdup(first_lang);
				}
			} else {
				DBG(DBG_MODNAME " Not a valid utf8 string: %s",
				    first_lang);;
			}
			voice->language = lang;
			voice->variant = variant;

			g_queue_push_tail(voice_list, voice);
		}

	}

	numvoices = g_queue_get_length(voice_list);
	DBG(DBG_MODNAME " %d voices total.", numvoices)

	if (PactosListVoiceVariants) {
		variant_spec = g_new0(espeak_VOICE, 1);
		variant_spec->languages = "variant";
		espeak_variants = espeak_ListVoices(variant_spec);
		variant_list = g_queue_new();

		for (i = 0; espeak_variants[i] != NULL; i++) {
			v = espeak_variants[i];

			vname = g_strdup(v->name);
			g_queue_push_tail(variant_list, vname);
		}

		numvariants = g_queue_get_length(variant_list);
		DBG(DBG_MODNAME " %d variants total.", numvariants)
	}

	totalvoices = (numvoices * numvariants) + numvoices;
	result = g_new0(SPDVoice *, totalvoices + 1);
	voice_list_iter = g_queue_peek_head_link(voice_list);

	for (i = 0; i < totalvoices; i++) {
		result[i] = voice_list_iter->data;

		if (variant_list && !g_queue_is_empty(variant_list)) {
			vo = voice_list_iter->data;
			variant_list_iter = g_queue_peek_head_link(variant_list);

			while (variant_list_iter != NULL && variant_list_iter->data != NULL) {
				voice = g_new0(SPDVoice, 1);

				voice->name = g_strdup_printf("%s+%s", vo->name,
							      (char *)variant_list_iter->data);
				voice->language = g_strdup(vo->language);
				voice->variant = g_strdup(vo->variant);

				result[++i] = voice;
				variant_list_iter = variant_list_iter->next;
			}
		}

		voice_list_iter = voice_list_iter->next;
	}

	if (voice_list != NULL)
		g_queue_free(voice_list);
	if (variant_list != NULL)
		g_queue_free_full(variant_list, (GDestroyNotify)g_free);
	if (variant_spec != NULL)
		g_free(variant_spec);

	result[i] = NULL;
	DBG(DBG_MODNAME " %d usable voices.", totalvoices);

	return result;
}

static void pactos_free_voice_list()
{
	if (pactos_voice_list != NULL) {
		int i;
		for (i = 0; pactos_voice_list[i] != NULL; i++) {
			g_free(pactos_voice_list[i]->name);
			g_free(pactos_voice_list[i]->language);
			g_free(pactos_voice_list[i]->variant);
			g_free(pactos_voice_list[i]);
		}
		g_free(pactos_voice_list);
		pactos_voice_list = NULL;
	}
}

static TPactosSuccess pactos_set_punctuation_list_from_utf8(const gchar * punct)
{
	TPactosSuccess result = ERROR;
	wchar_t *wc_punct =
	    (wchar_t *) g_convert(punct, -1, "wchar_t", "utf-8", NULL, NULL,
				  NULL);
	if (wc_punct != NULL) {
		espeak_ERROR ret = espeak_SetPunctuationList(wc_punct);
		if (ret == EE_OK)
			result = OK;
		g_free(wc_punct);
	}
	return result;
}

/* > */

/* local variables: */
/* folded-file: t */
/* c-basic-offset: 4 */
/* end: */
